# Copyright (c) 2024 Boris Soldatov
# SPDX-License-Identifier: MIT
#
# This file is part of an open source project licensed under the MIT License.
# See the LICENSE file in the project root for more information.

import asyncio
import os
import platform
import signal
from asyncio.exceptions import CancelledError

import uvicorn
import yaml
from fastapi import FastAPI
from shiva.lib.tools import Config
from loguru import logger

from shiva.common.daemon import Shiva
from shiva.common.logging import setup_loggers

# App init
app = FastAPI(debug=False)
app.logger = logger
app.shiva = None
config_path = None


# # Setup events
@app.on_event("shutdown")
async def shutdown_daemon():
    logger.info("Shutting down")
    await app.shiva.stop_async()
    logger.info("All instances stopped!")


@app.on_event("startup")
async def run_daemon(*args, **kwargs):
    logger.warning("Starting setup...")
    # loop = asyncio.get_running_loop()
    config_helper = Config(config_path)
    config = config_helper.get_config()
    # Logging
    setup_loggers(config)

    daemon = Shiva(config, None, app)
    app.shiva = daemon
    loop = asyncio.get_event_loop()
    daemon.loop = loop

    # Prepare daemon
    await daemon.prepare()

    # Run daemon
    loop.create_task(daemon.run())
    if platform.system() != "Windows":
        loop.add_signal_handler(signal.SIGINT, daemon.stop)
        loop.add_signal_handler(signal.SIGHUP, daemon.stop)
        loop.add_signal_handler(signal.SIGTERM, daemon.stop)


def main(config):
    try:
        import os
        import sys

        print(f"CWD: {os.getcwd()}")
        sys.path.append(os.getcwd())
        global config_path
        if config:
            config_path = config
        config_path = config_path or os.environ.get("SHIVA_CONFIG") or "./config.yml"

        # Get port/host if available
        config_helper = Config(config_path)
        config = config_helper.get_config()
        port = 8085
        host = "0.0.0.0"

        if type(config.get("common")) == dict:
            web = config["common"].get("web")
            if web:
                host = web.get("host", host)
                port = web.get("port", port)

        logger.info(f"Running web instance on: {host}:{port}")
        uvicorn.run("shiva.main:app", host=host, port=port, log_level="error")
    except CancelledError:
        logger.warning("Web application stopped!")



import importlib

import typer
from loguru import logger
from shiva import commands
from shiva.common.modules_helper import ModuleHelper
from shiva.const import MODULES_COMMANDS


class CommandHelper:
    def __init__(self):
        self.command = typer.Typer()
        import os
        print(os.getcwd())

    def _load_commands(self, mh_com, user_scope=False):
        objects = mh_com.get_libs(user_scope=user_scope)
        logger.info(f'Found {len(objects)} command packages. Discovering...')
        for cmd_root in objects:
            if hasattr(cmd_root, 'command'):
                name = None
                if hasattr(cmd_root, 'name'):
                    name = cmd_root.name
                    logger.info(f'Loading commands group: {cmd_root.name}')
                    self.command.add_typer(cmd_root.command, name=name, help=cmd_root.help_text)
                else:
                    logger.info(f'Loading single commands: {cmd_root.__name__}')
                    self.command.add_typer(cmd_root.command)
        logger.info('Shiva commands successfuly loaded.')

    def load_user(self):
        logger.info('Discovering users commands...')
        mh_com = ModuleHelper()
        try:
            module = importlib.import_module(MODULES_COMMANDS)
            mh_com.load_module(module)
            self._load_commands(mh_com, user_scope=True)
        except ModuleNotFoundError:
            logger.info(f"Can't find module '{MODULES_COMMANDS}' with users commands")

    def load_common(self):
        logger.info('Discovering shiva internal commands...')
        mh_com = ModuleHelper()
        mh_com.load_module(commands)
        self._load_commands(mh_com)



# Copyright (c) 2024 Boris Soldatov
# SPDX-License-Identifier: MIT
#
# This file is part of an open source project licensed under the MIT License.
# See the LICENSE file in the project root for more information.

from abc import ABC, abstractmethod
from collections import defaultdict

from loguru import logger

from shiva.common.config import ConfigHelper
from shiva.const import MODULES_WORKERS


class BaseDispatcher:
    base_class = None
    name = None
    load_workers = True

    def __init__(self, shiva, config, dispatcher_config={}):
        self.shiva = shiva
        self.config = config
        self.dispatcher_config = {'enabled': True, 'coro': 1, 'name': self.name, 'policy': 'ALL'}
        self.dispatcher_config.update(dispatcher_config)
        self.name = self.dispatcher_config['name']
        self.workers = defaultdict(list)
        self.coro = []
        self.workers_configurations = {}
        self.workers_all = {}
        self.workers = defaultdict(list)
        self.ch = ConfigHelper(self.config)
        self.policy_map = {
            'ALL': self.policy_all,
            'CONFIG': self.policy_config,
            'GROUP': None
        }
        self.policy = self.policy_map[self.dispatcher_config['policy']]

    async def prepare(self):
        pass

    async def load(self, scope):
        self.workers_scope = scope
        if self.load_workers:
            logger.info(f'[{self.dispatcher_config["name"]}]Preparing workers...')
            w_cls = scope.filter_members(self.base_class)
            for w in w_cls:
                # print(f'*********Trying to load [{w.name}] worker')
                w_dispatcher = None
                if hasattr(w, 'dispatcher'):
                    w_dispatcher = w.dispatcher
                # print(f'*********[{w.name}] dispatcher {w_dispatcher}: Current dispatcher: {self.dispatcher_config["name"]}')
                if w.name in self.workers_all:
                    logger.error(f'Worker already loaded: {w.name}. Check conflicting names.')
                if w_dispatcher and w_dispatcher == self.dispatcher_config['name']:
                    self.workers_all[w.name] = w
                    logger.info(f'Worker loaded: {w.name}')
                elif not w_dispatcher:
                    self.workers_all[w.name] = w
                    logger.info(f'Worker loaded: {w.name}')

            for w_name, cfg in self.ch.get_childs(MODULES_WORKERS, {'enabled': True}):
                if cfg.get('worker') in self.workers_all:
                    self.workers_configurations[w_name] = cfg
            logger.info(f'[{self.dispatcher_config["name"]}] Workers configuration loaded!')
            logger.info(f'[{self.dispatcher_config["name"]}]Loading workers(POLICY: {self.dispatcher_config["policy"]})...')
            self.policy()

    def policy_all(self):
        # Config first
        for w_sub_name, w_cfg in self.workers_configurations.items():
            dispatcher = self.is_compatible(w_cfg)
            if dispatcher:
                logger.warning(f'Loading {w_sub_name} to {dispatcher}')
                for w in range(w_cfg['coro']):
                    # self.workers[w_cfg['worker']].append(self.workers_all[w_cfg['worker']](self.shiva, self.config, w_cfg))
                    self.workers[w_cfg['name']].append(self.workers_all[w_cfg['worker']](self.shiva, self.config, w_cfg))
        # Other(not added from config)
        for w_name, w_cls in self.workers_all.items():
            if w_name not in self.workers:
                logger.warning(f'Loading {w_name} -> {self.name}')
                self.workers[w_name].append(w_cls(self.shiva, self.config))

    def policy_config(self):
        # Config first
        for w_sub_name, w_cfg in self.workers_configurations.items():
            dispatcher = self.is_compatible(w_cfg)
            if dispatcher:
                logger.warning(f'Loading {w_sub_name} to {dispatcher}')
                for w in range(w_cfg['coro']):
                    # self.workers[w_cfg['worker']].append(self.workers_all[w_cfg['worker']](self.shiva, self.config, w_cfg))
                    self.workers[w_cfg['name']].append(self.workers_all[w_cfg['worker']](self.shiva, self.config, w_cfg))

    def is_compatible(self, w_cfg):
        dispatcher = self.name
        logger.info(f'DISPATCHER DEFAULT: {dispatcher}')
        if hasattr(self.workers_all[w_cfg['worker']], 'dispatcher'):
            dispatcher = self.workers_all[w_cfg['worker']].dispatcher
        else:
            if w_cfg.get('dispatcher'):
                dispatcher = w_cfg['dispatcher']
        if dispatcher == self.dispatcher_config["name"]:
            return dispatcher

    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        raise NotImplementedError


class BaseWorker:
    def __init__(self, shiva, config, daemon_cfg=None):
        self.shiva = shiva
        self.config = config
        self.cfg_worker = daemon_cfg or {'enabled': True, 'coro': 1}

    async def prepare(self):
        pass

    async def start(self):
        pass

    async def stop(self):
        pass


class BaseMetric:
    def __init__(self, shiva, config, cfg_worker={}):
        self.shiva = shiva
        self.config = config
        self.running = False
        self.coro = []
        self.cfg_worker = cfg_worker or {'enabled': True, 'coro': 1, 'name': self.name}
        self.cfg_worker.update(cfg_worker)
        if cfg_worker.get('name'):
            self.name = self.cfg_worker['name']
        # print(self.cfg_worker)

    @abstractmethod
    async def prepare(self):
        pass

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass


class BaseDaemon:
    def __init__(self, shiva, config, cfg_worker={}):
        self.shiva = shiva
        self.config = config
        self.running = False
        self.coro = []
        self.cfg_worker = cfg_worker or {'enabled': True, 'coro': 1, 'name': self.name}
        self.cfg_worker.update(cfg_worker)
        if cfg_worker.get('name'):
            self.name = self.cfg_worker['name']
        # print(self.cfg_worker)

    @abstractmethod
    async def prepare(self):
        pass

    @abstractmethod
    async def start(self):
        pass

    @abstractmethod
    async def stop(self):
        pass


class BaseRmqWorker:
    load_globals = ('router', )

    def __init__(self, shiva, config, daemon_cfg=None):
        self.shiva = shiva
        self.config = config
        self.cfg_worker = daemon_cfg or {'enabled': True, 'coro': 1}

    async def prepare(self):
        pass

    async def run(self):
        pass

    async def start(self):
        pass

    async def stop(self):
        pass


class BaseProtocol:
    def __init__(self, shiva, workers):
        self.shiva = shiva
        self.workers = workers

    @abstractmethod
    async def prepare(self):
        pass

    async def start(self):
        pass

    async def stop(self):
        pass

    async def run(self, msg):
        pass



from abc import ABC, abstractmethod
import asyncio
from collections import defaultdict

from loguru import logger
from shiva.const import MODULES_DRIVERS


class BaseDriver(ABC):
    def __init__(self, config, shiva):
        self.config = config
        self.shiva = shiva

    @abstractmethod
    async def prepare(self):
        raise NotImplementedError

    @abstractmethod
    async def stop(self, message):
        raise NotImplementedError


class Connections:
    def __init__(self, shiva, config):
        self.shiva = shiva
        self.config = config
        self.drivers_all = {}
        self.connections = {}

    @staticmethod
    async def connection_prepare(task_connections):
        # task_connection_prepare = [task_connection.prepare() for task_connection in task_connections]
        task_connection_prepare = [asyncio.create_task(task_connection.prepare()) for task_connection in task_connections] # ESB-2359
        done, _ = await asyncio.wait(task_connection_prepare, return_when=asyncio.ALL_COMPLETED)
        for result in done:
            result.result()

    async def prepare(self):
        for d in self.shiva.scopes[MODULES_DRIVERS].filter_members(BaseDriver):
            self.drivers_all[d.name] = d
        set_connection_level = set()
        task_connections_level = defaultdict(list)
        for k, v in self.config.get('connections', {}).items():
            if v['driver'] in self.drivers_all:
                connection_level = v.get('level') or 0
                connection = self.drivers_all[v['driver']](v['config'], self.shiva)
                connection.connection_name = k
                set_connection_level.add(connection_level)
                task_connections_level[connection_level].append(connection)
                self.connections[k] = connection
            else:
                logger.error(f'Unable to load driver: {k}[{v["driver"]}]: Driver not found!')
        for connection_level in set_connection_level:
            logger.info(f"Running LEVEL={connection_level}")
            task_connections = task_connections_level[connection_level]
            for connection in task_connections:
                logger.warning(f'Adding driver (LEVEL={connection_level}): '
                               f'{connection.connection_name}[{connection.name}]...')
            await self.connection_prepare(task_connections)



import typer
# from art import tprint
from shiva.art import logo
from shiva.main import main

name = 'daemon'
help_text = 'Shiva common commands'
command = typer.Typer()


@command.command('')
def run(config: str = '', application: str = 'shiva_default'):
    """
        Run shiva daemon
    """
    # tprint('Shiva')
    print(logo())
    print('Starting shiva...')
    print(f'Starting shiva APP: {application} ...')
    main(config)



import typer

app = typer.Typer()


@app.command()
def word(word: str, second: str = 'default'):
    """
        Some help string
    """
    print(f'WORD: {word}')


@app.command()
def word_1(word: str):
    print(f'WORD: {word}')



from loguru import logger
from shiva.common.modules_helper import get_class_that_defined_method


class Router:
    def __init__(self):
        self.routes = {}
        self.func_names = set()

    def route(self, *args):
        def wrapper(fnc):
            if fnc.__name__ in self.func_names:
                logger.error(f'Function already loaded: {fnc.__name__}. Check conflicting names. (Class: {fnc.__qualname__.split(".")[0]})')

            for r in args:
                self.routes[r] = fnc.__name__
            self.func_names.add(fnc.__name__)
            return fnc
        return wrapper



import os
import sys

from shiva.common.commands import CommandHelper

ch = CommandHelper()
sys.path.append(os.getcwd())
ch.load_common()
ch.load_user()


def main():
    ch.command()


if __name__ == '__main__':
    ch.command()



from shiva.dispatchers import web


class DummyShiva:
    pass


def test_web_dispatcher_router_registration():
    shiva = DummyShiva()
    config = {}
    dispatcher_config = {"name": "dispatcher_web"}
    dispatcher = web.Web(shiva, config, dispatcher_config)
    assert dispatcher.name == "dispatcher_web"



import pymongo
from loguru import logger
from marshmallow import Schema, ValidationError, fields
from motor.motor_asyncio import AsyncIOMotorClient
from shiva.common.driver import BaseDriver


class Config(Schema):
    dsn = fields.String(required=True)
    pool_min = fields.Integer(required=False, default=1)
    pool_max = fields.Integer(required=False, default=50)
    indexes = fields.Dict(required=False)


class MongoDb(BaseDriver):
    name = 'mongodb'

    def validate(self):
        self.config = Config().load(self.config)

    async def prepare(self):
        """
        Init connection pool
        :return:
        """
        self.validate()
        self.pool = AsyncIOMotorClient(self.config['dsn'])

        # Indexes
        if self.config.get('indexes'):
            logger.info('Creating MongoDb indexes...')
            await self.create_indexes(self.config['indexes'])

        return self.pool

    async def create_indexes(self, indexes):
        for db_name, index_root in indexes.items():
            logger.info(f'Creating indexes for db: {db_name}')
            for collection_name, indexes_all in index_root.items():
                logger.info(f'Creating indexes for collection: {db_name}->{collection_name}')
                db = getattr(self.pool, db_name)
                collection = getattr(db, collection_name)
                for index_name, index_cfg in indexes_all.items():
                    indx = []
                    logger.info(f'Creating index [{db_name}->{collection_name}] {index_name}: {index_cfg}')
                    for indx_field, direction in index_cfg.items():
                        indx.append((indx_field, getattr(pymongo, direction)))
                    _ = await collection.create_index(indx, name=index_name)

    async def stop(self):
        logger.warning('Shutting down mongodb pool...')
        await self.pool.close()


if __name__ == '__main__':
    cfg = {
        # 'dsn': 'postgresql://postgres:@127.0.0.1:5432/trcont',
        'pool_min': 1,
        'pool_max': 50,
        'max_inactive_connection_lifetime': 300,
        'max_queries': 20
    }
    pg = MongoDb('p1', cfg)



import asyncio
import importlib
import os

import yaml
from loguru import logger

from shiva.common.driver import Connections
from shiva.common.modules_helper import Scope
from shiva.const import CLI_SCOPES, SHIVA_ROOT
from shiva.lib.tools import Config


class ShivaCLI:
    def __init__(self):
        self.config = self.get_config()
        self.scopes = {}
        self.connections = None
        self.croot = None

    @staticmethod
    def get_config():
        config_path = os.environ.get("SHIVA_CONFIG") or "./config.yml"
        config_helper = Config(config_path)
        config = config_helper.get_config()
        return config

    async def prepare(self):
        logger.info("Preparing...")
        self.load_scopes(CLI_SCOPES)
        await self.prepare_connections()

    async def prepare_connections(self):
        logger.info("Loading drivers...")
        self.croot = Connections(self, self.config)
        logger.info("Preparing drivers...")
        await self.croot.prepare()
        self.connections = self.croot.connections

    def _validate_scope_packages(self, packages):
        """
        Validate that scope packages are installed.

        Args:
            packages: List of package names to validate

        Returns:
            List of valid package names
        """
        valid_packages = []
        for package in packages:
            try:
                importlib.import_module(package)
                valid_packages.append(package)
                logger.info(f'Scope package "{package}" found')
            except ImportError:
                logger.warning(f'Scope package "{package}" not found, skipping')
        return valid_packages

    def load_scopes(self, scopes):
        logger.info("Loading shiva + packages + user scopes...")

        # Get additional packages from config
        scope_packages = self.config.get("scopes", {}).get("packages", [])
        valid_packages = self._validate_scope_packages(scope_packages)

        for scope in scopes:
            # Build scope loading chain
            # Order: shiva -> user -> packages
            sc_list = []

            # 1. Built-in shiva modules
            sc_list.append(f"{SHIVA_ROOT}.{scope}")

            # 2. User modules
            sc_list.append(scope)

            # 3. Modules from installed packages
            for package in valid_packages:
                sc_list.append(f"{package}.{scope}")

            # TODO: Current implementation loads first found module and skips duplicates.
            # If user modules need to override package modules, reverse the order.
            # See: shiva/common/modules_helper.py:Scope.load()

            logger.info(f'Scope "{scope}" loading order:')
            for idx, path in enumerate(sc_list, 1):
                logger.info(f"  {idx}. {path}")

            self.scopes[scope] = Scope(scope, sc_list)

        for name, scope in self.scopes.items():
            logger.info(f"{name}: {len(scope.scopes)} modules loaded")

    async def run(self):
        await self.prepare()
        await self.task()
        await self.stop_async()

    @staticmethod
    async def stop_async():
        import asyncio

        current_task = asyncio.current_task()
        tasks = [task for task in asyncio.all_tasks() if task is not current_task]
        for task in tasks:
            task.cancel()

    async def task(self):
        raise NotImplementedError



import orjson
from loguru import logger
from shiva.common.base import BaseProtocol


class JSON_UnroutedALL(BaseProtocol):
    name = 'JSON_UnroutedALL'

    async def prepare(self):
        self.routes = {}
        logger.info('Loading proto workers...')
        for w, inst in self.workers.items():
            if inst:
                self.routes[w] = inst[0]  # Workers list(we nead at least one)[<class>, ....]

    async def dispatch(self, msg):
        for w_name, w_obj in self.routes.items():
            logger.debug(f'[Sending msg to: {w_name}]')
            await w_obj.run(msg.body)



# Copyright (c) 2024 Boris Soldatov
# SPDX-License-Identifier: MIT
#
# This file is part of an open source project licensed under the MIT License.
# See the LICENSE file in the project root for more information.

SHIVA_ROOT = 'shiva'

MODULES_COMMANDS = 'commands'
MODULES_DRIVERS = 'drivers'
MODULES_WORKERS = 'workers'
MODULES_DISPATCHERS = 'dispatchers'
MODULES_TYPES = 'data_types'
MODULES_PROTO = 'proto'


SCOPES = [
    # MODULES_COMMANDS, Typer handling commands
    MODULES_DRIVERS,
    MODULES_WORKERS,
    MODULES_DISPATCHERS,
    MODULES_TYPES,
    MODULES_PROTO,
]

CLI_SCOPES = [
    MODULES_DRIVERS,
]

MSG_ACK = 0
MSG_NACK = 1
MSG_REJECT = 2



from shiva.proto.rmq_simple import JSON_UnroutedALL
from shiva.proto.bpm import RoutedPublisher
from unittest.mock import MagicMock, AsyncMock
import pytest


class DummyShiva:
    pass


class DummyWorker:
    async def run(self, body):
        self.last_body = body


@pytest.mark.asyncio
async def test_json_unroutedall_dispatch():
    shiva = DummyShiva()
    config = {}
    proto = JSON_UnroutedALL(shiva, config)
    worker = DummyWorker()
    proto.routes = {"w1": worker}

    class Msg:
        body = {"foo": "bar"}

    await proto.dispatch(Msg())
    assert worker.last_body == {"foo": "bar"}


@pytest.mark.asyncio
async def test_routed_publisher_publish():
    exchange = MagicMock()
    exchange.publish = AsyncMock()
    proto = RoutedPublisher(exchange, "system_from")
    await proto.publish("system_to", "object", "action", params={})
    exchange.publish.assert_awaited()



from aiobotocore.session import get_session
from marshmallow import Schema, INCLUDE, fields
from shiva.common.driver import BaseDriver


def client_context(fnc):
    async def wrapper(*args, **kwargs):
        self = args[0]
        async with self.session.create_client(
            endpoint_url=self.config["endpoint_url"],
            service_name=self.config.get("service_name", "s3"),
            aws_access_key_id=self.config["aws_access_key_id"],
            aws_secret_access_key=self.config["aws_secret_access_key"],
        ) as client:
            bucket = kwargs.get("bucket") or self.bucket
            if not bucket:
                raise Exception("Empty bucket")
            kwargs["bucket"] = bucket
            kwargs["client"] = client
            result = await fnc(*args, **kwargs)
            return result

    return wrapper


class Config(Schema):
    class Meta:
        unknown = INCLUDE

    endpoint_url = fields.String(required=True)
    aws_access_key_id = fields.String(required=True)
    aws_secret_access_key = fields.String(required=True)
    bucket = fields.String(required=True)


class s3Api(BaseDriver):
    name = "s3"

    def validate(self):
        self.config = Config().load(self.config)

    async def prepare(self):
        self.validate()
        self.session = get_session()
        self.bucket = self.config["bucket"]

    @client_context
    async def get_file(self, file_path, bucket, client):
        # res = await client.generate_presigned_url('get_object', Params={'Bucket': 'insurancedev', 'Key': 'bpm/test.pdf'}, ExpiresIn=600)
        # print(res)
        obj = await client.get_object(Bucket=bucket, Key=file_path)
        return await obj["Body"].read()

    @client_context
    async def put_file(self, file_path, body, bucket, client):
        result = await client.put_object(Bucket=bucket, Key=file_path, Body=body)
        return result

    @client_context
    async def publish_file_link(self, filename, expire, bucket, client):
        link = await client.generate_presigned_url(
            "get_object", Params={"Bucket": bucket, "Key": filename}, ExpiresIn=expire
        )
        return link

    @client_context
    async def delete_file(self, file_path, bucket, client):
        result = await client.delete_object(Bucket=bucket, Key=file_path)
        return result

    @client_context
    async def metadata_file(self, file_path, bucket, client):
        result = await client.head_object(Bucket=bucket, Key=file_path)
        return result

    async def stop(self):
        pass



import os

import yaml
from mako.template import Template


class Singleton(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Singleton, cls).__new__(cls)
        return cls.instance


class Config(Singleton):
    def __init__(self, config_path: str) -> None:
        self.path = config_path
        self.config = None

    def get_config(self) -> dict:
        if not self.config:
            with open(self.path, encoding="utf8") as f:
                cfg_raw = f.read()
                cfg_tmpl = Template(cfg_raw, strict_undefined=True)
                cfg_final = cfg_tmpl.render(**os.environ)
                self.config = yaml.load(cfg_final, Loader=yaml.SafeLoader)
        return self.config


import asyncpg
from loguru import logger
from marshmallow import Schema, ValidationError, fields
from shiva.common.driver import BaseDriver


class Config(Schema):
    dsn = fields.String(required=True)
    pool_min = fields.Integer(required=False)
    pool_max = fields.Integer(required=False)
    max_inactive_connection_lifetime = fields.Integer(required=False)
    max_queries = fields.Integer(required=False)


class Pg(BaseDriver):
    name = 'postgres'

    def validate(self):
        self.config = Config().load(self.config)

    async def prepare(self) -> asyncpg.pool.Pool:
        """
        Init connection pool
        :return:
        """
        self.validate()
        self.pool: asyncpg.pool.Pool = await asyncpg.create_pool(
            dsn=self.config['dsn'],
            min_size=self.config['pool_min'],
            max_size=self.config['pool_max'],
            max_inactive_connection_lifetime=self.config['max_inactive_connection_lifetime'])
        return self.pool

    async def stop(self):
        logger.warning
        await self.pool.close()


if __name__ == '__main__':
    cfg = {
        # 'dsn': 'postgresql://postgres:@127.0.0.1:5432/trcont',
        'pool_min': 1,
        'pool_max': 50,
        'max_inactive_connection_lifetime': 300,
        'max_queries': 20
    }
    pg = Pg('p1', cfg)



import asyncio

import aio_pika
from aio_pika.pool import Pool
from loguru import logger
from shiva.common.driver import BaseDriver


class RabbitMq(BaseDriver):
    name = 'rmq'

    async def prepare(self):
        self.loop = asyncio.get_running_loop()

    async def get_connection(self):
        return await aio_pika.connect_robust(self.config['dsn'])

    async def stop(self):
        pass




from fastapi import APIRouter
from fastapi_restful.inferring_router import InferringRouter
from loguru import logger

from shiva.common.base import BaseDispatcher, BaseWorker
from shiva.common.modules_helper import get_class_that_defined_method

# Black magic


class Web(BaseDispatcher):
    name = 'dispatcher_web'
    base_class = BaseWorker
    load_workers = False

    async def prepare(self):
        self.routers = []
        for s in self.workers_scope.scopes:
            if hasattr(s, 'router') and isinstance(s.router, (APIRouter, InferringRouter)):
                self.routers.append(s.router)
                # self.shiva.app.include_router(s.router)
        w_cls = self.workers_scope.filter_members(self.base_class)
        for w in w_cls:
            logger.info(f'WEB application loaded: {w.name}')
            if hasattr(w, 'call_prepare') and getattr(w, 'call_prepare') == True:
                inst = w()
                await inst.prepare()
        for router in self.routers:
            # router.shiva = self.shiva
            for r in router.routes:
                _cls = get_class_that_defined_method(r.endpoint)
                self.workers[f'[{_cls.name}]: {r.path}'] = [1, ]  # Hack for DispatchersRoot->map

    async def start(self):
        for router in self.routers:
            self.shiva.app.include_router(router)
            for r in router.routes:
                path = getattr(r, "path")
                logger.info(f'Web application started: {path}')

    async def stop(self):
        logger.warning(f'Stopped {self.name}')



import argparse


class ArgParse:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='Available commands:')
        self.args = None

    def parse(self):
        self.args = self.parser.parse_args()



import asyncio
import importlib
import sys
import time

from loguru import logger
from shiva.common.dispatcher import DispatchersRoot
from shiva.common.driver import Connections
from shiva.common.modules_helper import Scope
from shiva.const import SCOPES, SHIVA_ROOT


class Shiva:
    def __init__(self, config, loop, app):
        self.config = config
        self.loop = loop
        self.app = app
        self.running = False
        self.stopped = True
        self.scopes = {}
        self.dispatchers = None
        self.connections = None
        self.droot = None
        self.croot = None
        self.data_exch = {}

    async def prepare(self):
        logger.info('Preparing...')
        self.load_scopes(SCOPES)
        await self.prepare_connections()
        await self.prepare_dispatchers()

    async def prepare_connections(self):
        logger.info('Loading drivers...')
        self.croot = Connections(self, self.config)
        logger.info('Preparing drivers...')
        await self.croot.prepare()
        self.connections = self.croot.connections

    async def prepare_dispatchers(self):
        logger.info('Loading dispatchers...')
        self.droot = DispatchersRoot(self, self.config)
        logger.info('Preparing dispatchers...')
        await self.droot.prepare()

    def _validate_scope_packages(self, packages):
        """
        Validate that scope packages are installed.

        Args:
            packages: List of package names to validate

        Returns:
            List of valid package names
        """
        valid_packages = []
        for package in packages:
            try:
                importlib.import_module(package)
                valid_packages.append(package)
                logger.info(f'Scope package "{package}" found')
            except ImportError:
                logger.warning(f'Scope package "{package}" not found, skipping')
        return valid_packages

    def load_scopes(self, scopes):
        logger.info('Loading shiva + packages + user scopes...')

        # Get additional packages from config
        scope_packages = self.config.get('scopes', {}).get('packages', [])
        valid_packages = self._validate_scope_packages(scope_packages)

        for scope in scopes:
            # Build scope loading chain
            # Order: shiva -> packages -> user
            sc_list = []

            # 1. Built-in shiva modules
            sc_list.append(f'{SHIVA_ROOT}.{scope}')

            # 2. Modules from installed packages
            for package in valid_packages:
                sc_list.append(f'{package}.{scope}')

            # 3. User modules
            sc_list.append(scope)

            # TODO: Current implementation loads first found module and skips duplicates.
            # If user modules need to override package modules, reverse the order.
            # See: shiva/common/modules_helper.py:Scope.load()

            logger.info(f'Scope "{scope}" loading order:')
            for idx, path in enumerate(sc_list, 1):
                logger.info(f'  {idx}. {path}')

            self.scopes[scope] = Scope(scope, sc_list)

        for name, scope in self.scopes.items():
            logger.info(f'{name}: {len(scope.scopes)} modules loaded')

    async def wait_coro(self):
        logger.info('Coro waiter started!')
        await self.droot.start()
        if self.droot.coro:
            task = [asyncio.create_task(t) for t in self.droot.coro] # ESB-2359
            await asyncio.wait(task)
        # logger.warning('Root coro waiter stopped!')
        logger.error('Root coro waiter stopped!')
        self.running = False

    async def run(self):
        self.running = True
        logger.warning(f'Starting Shiva...[{self}]')

        self.loop.create_task(self.wait_coro())
        self.stopped = False
        while self.running:
            # logger.info('*' * 40)
            # logger.info(f'RUNNING: {self}')
            # logger.info('*' * 40)
            await asyncio.sleep(2)
        self.loop.create_task(self.stop_async())

    async def stop_async(self):
        logger.info('Stopping...')
        logger.info('Waiting for daemon...')
        # print(self.droot.dispatchers)
        for d_name, d in self.droot.dispatchers.items():
            logger.warning(f'Trying to stop: {d_name} instances...')
            for inst_name, inst_obj in d.items():
                logger.warning(f'Stopping: {d_name}->{inst_name}')
                await inst_obj.stop()
        current_task = asyncio.current_task()
        # print(current_task)
        # print('*' * 80)
        tasks = [task for task in asyncio.all_tasks() if task is not current_task]
        for task in tasks:
            # print(f'T: {task}')
            task.cancel()
            # print('>>CANCELED')
        # print('!!!!!!!GATHER!!!!!!')
        # await asyncio.gather(*tasks, return_exceptions=True)
        logger.info('Shiva stopped.')

    def stop(self):
        self.running = False
        logger.warning('Stop command received!')
        # while not self.stopped:
        #     logger.warning('Waiting for Shiva to stop...')
        #     time.sleep(1)
        # sys.exit(0)



from redis import asyncio as aioredis
from loguru import logger
from marshmallow import Schema, ValidationError, fields
from shiva.common.driver import BaseDriver


class Config(Schema):
    dsn = fields.String(required=True)
    minsize = fields.Integer(required=False)
    maxsize = fields.Integer(required=False)


class Redis(BaseDriver):
    name = 'redis'

    def validate(self):
        self.config = Config().load(self.config)

    async def prepare(self):
        """
        Init connection pool
        :return:
        """
        self.validate()
        dsn = self.config.pop('dsn')
        max_connections = self.config.get('maxsize', 10)
        connection_pool = aioredis.ConnectionPool.from_url(
            dsn,
            max_connections=max_connections
        )
        self.pool = aioredis.Redis(connection_pool=connection_pool)
        return self.pool

    async def stop(self):
        logger.warning("Closing Redis connection pool")
        await self.pool.aclose()



import asyncio
from collections import defaultdict

from loguru import logger
from shiva.common.base import BaseDispatcher
from shiva.common.config import ConfigHelper
from shiva.const import MODULES_DISPATCHERS, MODULES_WORKERS


class DispatchersRoot:
    def __init__(self, shiva, config):
        self.shiva = shiva
        self.config = config
        self.dispatchers_all = {}
        self.dispatchers = defaultdict(dict)
        self.configurations = {}
        self.coro = []
        self.ch = ConfigHelper(self.config)
        self.disabled_list = {}
        self.configure()

    def configure(self):
        for d_name, cfg in self.ch.get_childs(MODULES_DISPATCHERS, {'enabled': True}):
            self.configurations[d_name] = cfg
            # logger.error(f'Dispatcher {d_name}')
        logger.info('Dispatchers configuration loaded!')

    async def prepare(self):
        # If we have non default dispatchers - ignore default one
        non_default = []
        # Collect all dispatchers
        for d in self.shiva.scopes[MODULES_DISPATCHERS].filter_members(BaseDispatcher):
            self.dispatchers_all[d.name] = d
        # Config first
        for inst_name, cfg in self.configurations.items():
            if cfg['dispatcher'] in self.dispatchers_all:
                non_default.append(cfg['dispatcher'])
                self.dispatchers[cfg['dispatcher']][inst_name] = self.dispatchers_all[cfg['dispatcher']](self.shiva, self.config, cfg)
                logger.info(f'Loading dispatcher: {cfg["dispatcher"]}[{inst_name}]({self.dispatchers_all[cfg["dispatcher"]]})')
        logger.info('Dispatchers loaded.')
        logger.info('Loading workers...')
        for d_name, d_data in self.dispatchers.items():
            for d_sub_name, d_obj in d_data.items():
                await d_obj.load(self.shiva.scopes[MODULES_WORKERS])
                await d_obj.prepare()
                # Prepare workers:
                for w_name, w_list in d_obj.workers.items():
                    for w in w_list:
                        if hasattr(w, 'prepare'):
                            await w.prepare()
        self.get_map()

    def get_map(self):
        for name, d in self.dispatchers.items():
            print(f'Dispatcher: {name}:')
            for d_name, d_obj in d.items():
                print(f'\tName: {d_name}')
                print(f'\tPolicy: {d_obj.dispatcher_config["policy"]}')
                print('\t\tWorkers:')
                for w_name, w_obj in d_obj.workers.items():
                    print(f'\t\t\t - {w_name}')
                    print(f'\t\t\t     coro: {len(w_obj)}')

    async def start(self):
        for d_name, d_inst in self.dispatchers.items():
            logger.info(f'Starting [{d_name}] instances...')
            for d_inst_name, d_obj in d_inst.items():
                logger.info(f'Starting [{d_inst_name}]...')
                await d_obj.start()
                self.coro += d_obj.coro
        # for name, dispatcher in self.dispatchers_all.items():
        #     logger.info(f'Starting {name}')
        #     self.coro.append(dispatcher.start())



import logging
import sys

import sentry_sdk
from loguru import logger
from sentry_sdk import set_level, utils
from sentry_sdk.integrations.logging import BreadcrumbHandler, EventHandler, LoggingIntegration


def setup_loggers(config):
    logger.remove()
    logger.add(sys.stdout, colorize=True, level=config["logging"]["level"])

    if config["logging"].get("sentry"):
        sentry_env = "DEFAULT"
        if config["logging"]["sentry"].get("environment"):
            sentry_env = config["logging"]["sentry"]["environment"]
        logger.error(f"Sentry env: {sentry_env}")
        sentry_sdk.init(
            config["logging"].get("sentry", {}).get("dsn"),
            traces_sample_rate=0,
            max_breadcrumbs=100,
            debug=False,
            attach_stacktrace=True,
            include_local_variables=True,
            environment=sentry_env,
            max_value_length=1000,
        )

        # Sentry integration crunch
        set_level(config["logging"]["level"].lower())
        # utils.MAX_STRING_LENGTH = 1000

        _ = logger.add(
            BreadcrumbHandler(level=logging.DEBUG),
            level=logging.DEBUG,
        )

        _ = logger.add(
            EventHandler(level=logging.ERROR),
            level=logging.ERROR,
        )

        integrations = [
            LoggingIntegration(level=None, event_level=None),
        ]
        logger.debug("Sentry enabled")



import random

logo_0_120= """
                                                                                                                        
                                                                                                                        
                                                                                                                        
                                                           .+.                                                          
                                                  `.      /oos/`      .`                                                
                                         `.      ./++-`  `+ssso`   `:+o/.                                               
                                        `-+:/.  `//oo+    :ooo/    -oo++-   `.-:/                                       
                                 .      `/++/:   .oys+:://+yss/::---sso/`   -+oo+.                                      
                                `///:.  `-+yso+oossyyyssssssooss++++/+o+/:-.+o++/`  `.:-:                               
                                `/+o+/`.:+syhyyyyyyyyhyhhhdhhhhhhhyyssoo+++/++o/.`  :oo++`                              
                         .`.``   -/ssysyyyyyyhhhyyyyyyssyoyyyysos+ssshyyyyyso++/+/:-so+/:    `                          
                         -+:/-. `:oyyyyyyhhyysoo/:-.....yhhhyhyys-....-:/+osyyyyso+++o+-   -/+:+`                       
                         ./+o+++syyyhhhyyso/-..`````    /yyhyyyys`    .::::++oooyyhyo++++--ooo/+`                       
                   ` ```  .-oyyyyhhyyso/-``..-.``.``   ``+ys+osss`    ````.-:-/oo+oyyhysoosss+/-                        
                   -+//:. .+ysyhhyso/.` ..-`  `````.` :++osyhyyyo/.` ```````  `:.+//+syhhsoos/`  .-:..`                 
                   `/+oo++ssyyhys+-`  `-. ````...`.`..:oyooso+osso/` ````````     ::`-+syhhysso--+oo/+                  
                    `-:sysyhhys+.     :-`````.`````..-///+oso/:oso/.`````  `````````. `-/syhhhsysyo++.                  
              ``.``  `/ssyhhs+-      --`````-`..```..:/o+syssooshso..``````.```     .`  `-+soo+syy+.`                   
              .://+/.ossyhhs:`       :/o-.`.-``.`.```.+hyoosoysyysy-````````  .```````    `/oooshyy/ .://`.             
              `:/+oyssshhyo-        .ooos+` .``.`.``.--/hossosyoso:````````.```  ````   .:++ooyhdhyy+o/o+o.             
               `..:ysyhhyo.        `+oo+/.``.:..` ``.`.-sddhysyhhy-```.`````  `````   `:oo+/:/oshhhyyyoo/.              
             ``   osyhhyo.         -oos. ```/ys/`..-://+hhdddysyyhs//++:-.`         `:oyo-    .shhhhyy/                 
          `.:+::-:sshhhs-           -syy+.`.oyyyooooysssyhyhhyyyyyhhoossyoo/:-   `./oss/`      .syhdhhy.`.::`           
           -+/oosyshhhs/             .oyhhsoso+sos/+ooooshysohhssyo//yhyyo/+/+///+oos+.         -syhdhhss/o+/-`         
             ---oyyhdys`              `+hyoyhs+oo++oyoshoss/+ss++/-/sss+-//+oo/ossyo:`           +shhdhhyss+:           
                ssyyhy/             `-/ososoososo/:.  .+sss//oo/o+oo.`.`    .-/+o/-`             .syhdyhs```            
           .-..-ysosyy:          `:+sso:../+/:--`    `-os+---+oyyyy:`````                         +sdhhhh.````          
         `:o/ooyyyhyyyy.       .+ss+-`            `:oyhysosydddhyyyooooooo/.                      :shhhhho/+o/.`        
         `.//++oyhdhyyys-     -ss/`            `:+yyss+/:+yhhhhysyy:` `-:oyyo-                    .shhhhhhsso+.         
             ` +syhdhy+sso:..:yy:            `+oso/-.     /yyhho-/ys+/-   .+yso:`                 .yyhhhh+.:`           
               /yyhdyo `-/osso/.            `oo/-        `+syhhy+-sh+syo.   .yhys+:-`             .syhhhh/              
          `:///syyhhhs`                    -oo:          -yhyyyhyyhyssshy.   `shshysss+::--.`     -yhhhhh:-/-`          
         `-oooosyyyhhs-                   /s-.           :yhhhhhyssssoosys    `oyo/:/+osoossso/`  /shhhyhs+oos-`        
           ..:. syyhhy/                   .o          `.-::/ss+shhdhhhhshy:     -sso+/:-``.-/syy+.syhhhyo/+s+.          
                :yyhdhs`                           `-/oo+oooooo+o+ooosssy+        ./osyyys/`  /yyyydhhhy-  `            
             .///hyyhhy+                 `--://::/+oo++++ossyyo+::/+++/+/`            `-+yys-  /hhyhdhyy-::             
            .oooohyyhhhy:           `:/osyyhhyssoo+++ohddddmmhy:   `                     .yyy: `hhhdhyhyooos.           
            .///+.+syhdhy-           osoyyyyo+-`    .ydhoydhys+`                          `yys:/sddhyy:/+o+..           
                   oyyhhhs-          `//+/-`       `ohy/yyso-                              -yyyyhdhyy/                  
                `::+hyyyhhy:                       /syoooo:`                               .hhhhhhyyo::-                
                /soossyyyhhh+`                    -osyo/o-                                 .yhhhhyyhsoos.               
               .:+++:`/yyyhhhs-                   `+o+//ss:                               :syhhhys-.+/o:-               
                      `/hyyshhh+.                   -sooyyyo`                           -+yyyhyyo.                      
                    `+ooohyysyhhy+.                  .ssssss+                         -/syhhyyyy+++`                    
                    .osss/-oysoyhh+/:`                .sys+os`                     `:oyhhhyys/oysss.                    
                    ``-`.  `:sysssss+++:.              `/yoos:                  .-/oyyhhyss/. `:.:`.                    
                           :+syhysss/://oo:.`            :yssy.             .-/osyoshyyyyhso/`                          
                          `+sss+./syyo+:+/sso+/-.`     `-/o+ysy/`      `.-/oossyyhyyss+-oysys.                          
                           :.:-`  `./oyso//++hhysso////oysyhhhhh:--./oosssssyssyyyy+:`  .-///`                          
                                     `.:+oyyydyoyyyshmhyhdhhhdhyyhhyhyhyyyyysoys+:`                                     
                                         `/smdyyysyhdmdhhys:o+syhyshhyshhdhs+://                                        
                                           -hmdmddddmddmyosyhsyhds/yhyyhhy+/-                                           
                                          -+dhddhysyhddyyhmmmddmNho+hyhdhdhhy-                                          
                                         .syyyhdddddhhddmmmmmdddmmdhddddddyyys`                                         
                                          :syyyhhyyydhysshhhsysyhyyyyssshhyhs:                                          
                                     `:::://sddmdhhhmhhosdhhoohdyyyhmhyhdmdh+::::-                                      
                                     -++sysosdmmmmmmmmmddmmmdddmddddmmmddmmhysso++`                                     
                                     /oyhyyyyhdddddddddddddddddddddhhhhdddddhhhdyo:                                     
                                    ./+yhhyhhhyhhhhhhhhhhhhhyyyssssyyysyhddhdddhso+`                                    
                                   `/+++oshddddhdhhhhhysshhyhs+//s:so//hhmmddyso++++`                                   
                                   :oooooosyhhhdhhhhhhyyyyhhhyoosysyyssyhhhhsooooo+o/                                   
                                  `+oyddddddddhhhhhhhhhdhhhhhhhhhhhyhhyyhyhyhhhhhhyoo`                                  
                                  :ooyhhhhhhhhhhhhhhhhyhyyyyhyhyyhhyyyssyyyyhhyhyysoo:                                  
                                 `ooshhhhhhhhhddddddhhyyyyyyyyyyyyyyyyyysssssyyyyyyooo                                  
                                 /ssyyyyysysyhhyyysoo+s+++/////:+/o/::-::--//-+//////s-                                 
                                .yhhhhhhhhhhhhyhyyyyyyysssssssssysysooo+oooososooooooo+`                                
                               `-+++oooooooooooooooooooo++++++++++++++++++++++++++++///.`                               
                                ``...................................................```                                
                                                                                                                        
"""

logo_0_80 = """
                                                                                
                                       `:`                                      
                                 .-`  .oo+`  `.-                                
                           `:-` `/++  `oo+   +o+.  `.-                          
                      ..`  :++:.-+yo/++so+/::+o+.``+++`  ``                     
                     `/+/..:syyyyyyyyyhhyyyysoo++//oo-` :++.                    
                `--.  -syyyyyyhyssoo+syyyys+++ossysso+/:s+:` `..                
                 /++:/syyyyso/:-..`  /hyyyy:  `-:/+ssyys+o/.-+++`               
             ... `/yyyhys+-..`.``.` `-ososs:   ``.-.:+ooyysoos/. ``             
             /+o/+syhy+-` .`.```````/sosssso- ````````-:-+syyso--///            
              .osyhy+.   -.``..````-/+oso+ss:````````  `` .+yhssss/-            
         `::/.+syhs-    `/:.`.`.```.oyososyso.````` ````.   :++sys.`--`         
          :/osshho`     /oo+`.`.`...-sssoyss-```````` ``` ./++ohdyyo+o:         
         ` `+shho`     .os-``-o/`.-::sddhyhy/:/:.`      `/o/.`.ohhhy/.          
       .//:/shhs`       -sy/-+ssooososhyyysyyosys+/:``./oo-    `ohhhs-:/.`      
        -:/yyhy:         .shsysoo/osysyooyo+:oso///++oso:`      .shdhyo+:`      
        ` -ssys`       `:+++oo+o/-` -so/:ooos-``  `.:/-`         +hhhy``        
      `///syyyy-     :+o:.  `.`` `-+yy+oyhyyy////:`              -ydhh/:/.      
      `-::oyhhys:` `+s-        -/os+:-/yhhsss-`.-+s/.            `yhhhso/.      
          oshy+-/o+o+.       `/+:`    .syhs:yos/` .sy+:.`        `shhh-         
       :/+yyhh+             .+/`      /hhyhsysoys` `oyyoso/:::.  .yhhh:/:`      
      `-:::yyhs`            /-      `./+ssyhhhysh/   :so//:-:+yo-/yhhsoo+.      
         ``syhh:              .``.-/ooosso+/++oss`     -/+ss/``ohyhhy:``        
        -ooyyhhy.       `-:+sysoo+++syhhdy:`.-.`           -ys.`hhhyyoo+`       
        `::./yhhs`      .ssyo/:`  `shsyyo:                  -ys/ydhs-:/.`       
           .-yyhhs.       ``      /yooo-                     shhhhy/-           
          .ooooyyhy:             .os+o-                     .shhhssos:          
           ````/yyyho.            .o+sy+`                 `:syhyo```.           
             `oossysyy+.           `osss:               `:oyhysyoo`             
             `-:-`-sssso+:-`        `/soo            `-/shhys:`//:.             
                  /oyoss+//+/-.       +os+`      `-:+syshyohso`                 
                  :/:``:+so+++ys+/--:osshhs..-/+oosssyyo/.`/++                  
                         `-/shhsyyymhhhsyyyyyyyyyyyos/.`                        
                            .yddhhdddhosssdyohyyhs/.`                           
                            /yhddyhhdydmmdmdshhddhh/                            
                           `+syhhyhhyhdhyyhhyhysdyy+                            
                         :/++odmmhddhyddshdhhdhhddy//:-                         
                        `+yysyddddddddddddddddddddhhhy+                         
                        -+syhhhhhhhhyyhhysososoyhddhhs+:                        
                       .+oooyhddhhhhyyhhho+soyoyhdhsoooo.                       
                       /ohdhdhhhhhhhhhhhhhhhyyyyyyyhhhyo+                       
                      .oshhhhhhhhhhhyyyyyyyyyyyssssyyyyso.                      
                      oyhyyyyyhyyyssso+++++oo+///://o///+/                      
                     .osssssssoooooooooooooooooo++oooooo+/`                     
                      `````..........```........``````````                      
"""

logo_1_80 = """
                                     -+`                                        
                                    .-ho                                        
                               .::` `+ho.                                       
                                 /. :shdy/`                                     
                             `/::  `:/yddd+`                                    
                               .. `./:-:-`hy`                                   
                               `+:--/s/syhddo                                   
                               `-:``.+:-+hydh                                   
                               `+o`:o.os+osds                                   
                               .-`   .ydddddy                                   
                               .:o/ ``sddddddy/`                                
                               :sys  `/o/sddddd/                                
                            .+.`/d+     /dddys:                                 
                           -/:+s.h`     /y./:s.                                 
                           `/.`-:ys/`   `/`. -y- `-oo-                          
                      `.       `/-+sh+/++/::--.:-   hdo                         
                     `.   `-    .`  -/:::-:/  `yy  `hdd.                        
                     ..  .+ds   ```-`    -:-  `hdy:/hdds.                       
                     .-  :ddy`  .. `.:-.`-:. `/ddd+  :hdd-                      
                    ``..`.dddh:. `...`.-.`/.-odddd..-:/dds                      
                    .//::hdhd.   .`   .:  +  .dddh````.ddy                      
                       `sdo`y:    .-     ::  :dddh    +ddh`                     
                     `:hh-  .h-    `..-/--   +ddd:     /ddy`                    
                     `hd/    .s:      .+     sddy       `-+y                    
                     .ds     /o`      `-     .odd-         ..                   
                    `sy`    :d:        :      oddd:      `..-`                  
               `-..`+:     :ddh+.          `:oydhhy ` ``.`-..`/`                
              `--`-s`.:-::/hd+-...   `-::::--:/.`/oy+yh+`-  ``:.                
              `..  o`  ``.:+s/-.         -::-.::`./--/`  ````-/+                
             `-````/ `.-:+oyysoo/`     `//+shy-  ..       `.../.                
             .`..```        .+hddyoossyso//--+`          -o::/.                 
             `.``.`       `..`:--..``               `-+shddy`                   
               `` `.`  ``..-----::://:-.       `-/oyddho+:.                     
                    `-/-:------....`      `-/oyddddddd.                         
                     :s--     -::--:/+++/:::--. .:osys/                         
                      ::.     `                                                 
                                                                                
                                                                                
"""

logo_2_80 = """
                                       .                                        
                                     -NMd`                                      
                                     hMMMo                                      
                                     NMMMy                                      
                                    -MMMMh                                      
                                    yMMMMM+                                     
                                    /MMMMm:                                     
                   `yh`              yMMN-`                                     
                   `NMN.       -+shdNMMMMMMNhs/`          .s:                   
                   .M.       +mMMMMMMMMMMMMMMMMm:      `:omN:                   
                   -mNds:`-yNMMNydMMMMMMMMMNsyNMMh- `/smshMNo`                  
                   :NmhNMMMMms:-mMMMMMMMMMMMMd./hMMmMMNyss``                    
                  `oNMdymy+- .yMMMh+mMMMMMohMMMs`.yMMdsdMm/                     
                   `.`  .sNNhMMm+`  /MMMMMm-`+mMNNMd/   ``                      
                          `ods- .+hNMMMMMMMN.  -ss-                             
                             :yNMMMdhMMMMMMMy                                   
                        ..:yNMMmy+..hMMMMMMMh                                   
                    /mmNMMNs+-`   -NMMMMm:-.                                    
                     .:-..`      :NMMMMh.                                       
                                :NMMMy.                                         
                               -NMMh-                                           
                               oMMN:                                            
                                sMMM/                                           
                                 oMMMo                                          
                                  :mMMo                                         
                                   `oMMo                                        
                                  .++hmNy                                       
                                   ``                                           
                                                                                
"""


def logo():
    logos = (logo_0_80, logo_1_80, logo_2_80)
    return logos[random.randint(0, len(logos) - 1)]



import typer
from loguru import logger

name = 'project'
help_text = 'Shiva project helper commands'
command = typer.Typer()


@command.command('')
def init():
    """
        Init shiva project
        Create base folders
    """
    logger.info('Initialize started...')



import datetime
import uuid
from collections import defaultdict
from copy import deepcopy

import orjson
from aio_pika import Message
from loguru import logger
from shiva.common.base import BaseProtocol
from shiva.const import MSG_ACK, MSG_NACK, MSG_REJECT

# {
#     "action": "all",
#     "object": "location",
#     "params": {
#         "data": [
#             {
#                 "id": "T_CN_CAN_CRR",
#                 "name": "Гуанчжоу (Guangzhou), депо CRRC",
#                 "type": "Терминал",
#                 "subtype": "Депо",
#                 "depo_region": "AE_CN_CAN"
#             }
#         ]
#     },
#     "priority": 1,
#     "system_to": "bpm",
#     "message_id": "de897640-1439-4eeb-a2a9-da14f081f55e",
#     "api_version": "1.0",
#     "system_from": "catalog",
#     "reference_id": "71967781",
#     "datetime_created": "2021-03-10 11:00:26.816493"
# }

MSG_TEMPLATE = {
    "system_from": None,
    "system_to": None,
    "object": None,
    "action": None,
    "reference_id": None,
    "message_id": None,
    "params": {
    },
    "priority": 1,
    "api_version": "1.0",
    "datetime_created": None
}


MSG_REPLY_TEMPLATE = {
    "system_from": None,
    "system_to": None,
    "object": None,
    "action": None,
    "reference_id": None,
    "message_id": None,
    "params": {
        'status': {
            'success': True,
            'request_id': None,
            'errors': []
        },
    },
    "priority": 1,
    "api_version": "1.0",
    "datetime_created": None
}


class RoutedPublisher:
    def __init__(self, exchange, system_from, msg_source={}):
        self.exchange = exchange
        self.system_from = system_from
        self.msg_source = msg_source

    async def publish(self, system_to, m_object, action, params={}, reference_id=None, route_key=None, success=True, errors=[]):
        if not route_key:
            route_key = f'{m_object}.{action}'
        msg_id = uuid.uuid4()

        if not reference_id:
            reference_id = msg_id
        data = deepcopy(MSG_TEMPLATE)
        data['system_from'] = self.system_from
        data['system_to'] = system_to
        data['message_id'] = msg_id
        data['reference_id'] = reference_id
        data['datetime_created'] = datetime.datetime.utcnow().isoformat(timespec='seconds')
        data['object'] = m_object
        data['action'] = action
        data['params'].update(params)
        await self.exchange.publish(Message(orjson.dumps(data)), route_key)
        logger.info(f'[publisher] Message sended: {route_key}')

    async def publish_reply(self, msg, success=True, errors=[]):
        route_key = f'{self.msg_source["object"]}.{self.msg_source["action"]}_reply'
        msg_id = uuid.uuid4()
        data = {
            "system_from": self.msg_source['system_to'],
            "system_to": self.msg_source['system_from'],
            "object": self.msg_source['object'],
            "action": f'{self.msg_source["action"]}_reply',
            "reference_id": self.msg_source['reference_id'],
            "message_id": msg_id,
            "params": {
                'status': {
                    'success': success,
                    'request_id': self.msg_source['message_id'],
                    'errors': errors
                },
            },
            "priority": 1,
            "api_version": "1.0",
            "datetime_created": datetime.datetime.utcnow().isoformat(timespec='seconds')
        }
        if msg:
            data['params'].update(msg)

        # if self.msg_source.get('reference_id'):
        #     data['reference_id'] = self.msg_source['reference_id']
        # else:
        #     data['reference_id'] = msg_id
            
        # if self.msg_source.get('message_id'):
        #     data['params']['status']['request_id'] = self.msg_source['message_id']

        await self.exchange.publish(Message(orjson.dumps(data)), route_key)
        logger.info(f'[publisher_reply] Message sended: {route_key}')


class JSON_Routed_BPM(BaseProtocol):
    def __init__(self, *args, exchanges=None):
        super().__init__(*args)
        self.exchanges = exchanges
        self.routes = {}

    async def prepare(self):
        routes = defaultdict(list)

        # TODO: Handle with responces and exchanges
        for k, v in self.exchanges.items():
            self.default_exchange = v
            break

        for w, inst_all in self.workers.items():
            for i in inst_all:
                if i:
                    # i = inst[0]
                    i.exchanges = self.exchanges

                    # Default exchange
                    i.default_exchange = self.default_exchange
                    # routes:
                    if hasattr(i, 'router'):
                        for r_name, fnc_name in i.router.routes.items():
                            routes[r_name].append(getattr(i, fnc_name))
                            logger.info(f'Routing loaded: {r_name}')
                    else:
                        logger.error(f'[JSON_Routed_BPM] class: {i.__class__.__name__} router REQUIRED!')
        for k, l in routes.items():
            self.routes[k] = tuple(l)

        # TODO: Handle with responces and exchanges
        for k, v in self.exchanges.items():
            self.default_exchange = v
            break

    async def reply(self, msg, msg_src):
        route_key = f'{msg_src["object"]}.{msg_src["action"]}_reply'
        publisher = RoutedPublisher(self.default_exchange, msg_src['system_to'], msg_src)
        await publisher.publish_reply(msg)

        # async def publish_reply(self, system_to, m_object, action, params={}, route_key=None, success=True, errors=[]):
        # msg_id = uuid.uuid4()
        # data = {
        #     "system_from": msg_src['system_to'],
        #     "system_to": msg_src['system_from'],
        #     "object": msg_src['object'],
        #     "action": f'{msg_src["action"]}_reply',
        #     "reference_id": msg_src['message_id'],
        #     "message_id": msg_id,
        #     "params": {
        #         'status': {
        #             'success': True,
        #             'request_id': None,
        #             'errors': []
        #         },
        #     },
        #     "priority": 1,
        #     "api_version": "1.0",
        #     "datetime_created": datetime.datetime.utcnow().isoformat(timespec='seconds')
        # }
        # if msg:
        #     data['params'].update(msg)
        # await self.default_exchange.publish(Message(orjson.dumps(data)), route_key)
        logger.info(f'Message sended: {route_key}')

    async def on_error(self, msg_raw, error):
        pass

    async def dispatch(self, msg):
        msg_json = orjson.loads(msg.body)
        p_key = self.get_p_key(msg_json)
        logger.info(f'Dispatching[{self.__class__}] {p_key}')
        # logger.info(self.routes)
        for f in self.routes.get(p_key, []):
            try:
                reply = await f(msg_json['params'])
                if reply:
                    await self.reply(reply, msg_json)
            except Exception:
                logger.exception(f'Unable to process message: {p_key}: {f}')

    def get_p_key(self, msg):
        p_key = None
        try:
            p_key = f'{msg["object"]}.{msg["action"]}'
        except Exception:
            logger.exception(f'Unable to get p_key: {msg}')
        return p_key



from loguru import logger


class ConfigHelper:
    def __init__(self, config):
        self.config = config

    def get_childs(self, root_key, filters={"enabled": True}):
        data = []
        for child_key, child_data in self.config.get(root_key, {}).items():
            filters_success = True
            for k, v in filters.items():
                d_key = f"{root_key}:{child_key}:{k}"
                # print(f'{d_key} -> {self.config[d_key]}')
                if self.config.get(root_key, {}).get(child_key, {}).get(k) != v:
                    filters_success = False
                    break
            if filters_success:
                data.append((child_key, self.config[root_key][child_key]))
        return data

    def gen_filters(self, filters):
        pass



import shiva.shiva_cli


def test_cli_main(monkeypatch):
    called = {}

    def fake_command():
        called["yes"] = True

    monkeypatch.setattr(shiva.shiva_cli.ch, "command", fake_command)
    shiva.shiva_cli.main()
    assert called["yes"]



import asyncio

from loguru import logger
from shiva.common.base import BaseDaemon


class MyDaemon(BaseDaemon):
    name = 'waiter_daemon'

    async def prepare(self):
        pass

    async def start(self):
        logger.info('Waiter started!')
        self.running = True
        while self.running:
            # logger.info('Waiter sleep...')
            await asyncio.sleep(5)
        logger.warning(f'Stopped: {self.name}')

    async def stop(self):
        self.running = False



# Copyright (c) 2024 Boris Soldatov
# SPDX-License-Identifier: MIT
#
# This file is part of an open source project licensed under the MIT License.
# See the LICENSE file in the project root for more information.




# import asyncio
# import platform

# import uvicorn
# from shiva import app

# if platform.system() != 'Windows':
#     import uvloop


# # UVLOOP
# loop = asyncio.get_event_loop()
# asyncio.set_event_loop(loop)

# if platform.system() != 'Windows':
#     uvloop.install()

import re
import sys

from shiva.shiva_cli import main

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(main())



import asyncio
import datetime

from loguru import logger
from shiva.common.base import BaseRmqWorker
from shiva.common.rmq_routing import Router

router = Router()


class RmqBench(BaseRmqWorker):
    name = 'shiva_benchmark'
    dispatcher = 'shiva_bench'

    async def prepare(self):
        pass

    async def start(self):
        pass
        # print('&' * 100)
        # print('&' * 100)
        # print(self.exchanges)
        # print('&' * 100)

    @router.route('shiva.bench')
    async def bench(self, message, raw=None):
        logger.info('Bench.')
        # await asyncio.sleep(3)



from fastapi import APIRouter
from loguru import logger
from shiva.common.base import BaseWorker
from shiva.dispatchers.web import Web
from shiva.main import app

router = APIRouter(prefix='')
shiva = None


class Healthcheck(BaseWorker):
    name = 'healthcheck'
    dispatcher = Web

    @router.get("/readiness")
    def read():
        # print('*' * 80)
        # print('*' * 80)
        # print('*' * 80)
        # print(app.shiva)
        status = {
            'status': 'ok'
            # 'running': app.daemon.running,
            # 'channels': len(app.daemon.channels),
        }
        logger.debug('Healthcheck ok')
        return status

    @router.get("/liveness")
    def live():
        status = {
            'status': 'ok'
            # 'running': app.daemon.running,
            # 'channels': len(app.daemon.channels),
        }
        # logger.error('Daemon test!')
        # raise Exception('Some error')
        logger.debug('Liveness ok')
        return status



import asyncio
import time

from aio_pika import Message
from aio_pika.exceptions import ChannelClosed, ChannelInvalidStateError, QueueEmpty
from aio_pika.pool import Pool
from loguru import logger

from shiva.common.base import BaseDispatcher, BaseProtocol, BaseRmqWorker
from shiva.const import MODULES_WORKERS, MSG_ACK, MSG_NACK, MSG_REJECT


class RmqDispatcher(BaseDispatcher):
    name = "rmq"
    base_class = BaseRmqWorker

    # async def load(self, workers_scope):

    async def prepare(self):
        self.running = False
        self.exchanges = {}
        self.queues = {}
        self.channels = []
        self.queue_iters = []
        self.connection_class = self.shiva.connections[self.dispatcher_config["connection"]]
        self.proto = None
        logger.warning(f"Preparing: {self.name}...")
        logger.info(f"Connection: {self.dispatcher_config['connection']}: {self.connection_class}")
        self.connection = await self.connection_class.get_connection()
        logger.info("Connected!")
        await self.init_exchanges()
        await self.load_proto()

    async def load_proto(self):
        for c in self.shiva.scopes["proto"].filter_members(BaseProtocol):
            if c.__name__ == self.dispatcher_config["proto"]:
                self.proto = c(self.shiva, self.workers, exchanges=self.exchanges)
                await self.proto.prepare()
                logger.info(
                    f"[{self.__class__.__name__}][{self.dispatcher_config['name']}] Loading protocol: {c.__name__}"
                )

    async def init_exchanges(self):
        logger.info("Init exchanges...")
        for exch_name, config in self.dispatcher_config["config"]["exchanges"].items():
            logger.info(f"Configuring exchange: {exch_name}")
            # V2
            v2_enabled = False
            v2_system = None
            if config.get("routing"):
                if config["routing"].get("v2_enabled") and config["routing"].get("system"):
                    v2_enabled = config["routing"]["v2_enabled"]
                    v2_system = config["routing"]["system"]

            # Exchange only channel
            exch_chan = await self.connection.channel(**config.get("config_channel", {}))
            self.channels.append(exch_chan)
            exchange = await exch_chan.declare_exchange(exch_name, **config["config"])
            self.exchanges[exch_name] = exchange
            if config.get("queues"):
                for q_name, q_cfg in config["queues"].items():
                    self.queues[q_name] = []
                    # Separate channels for queues
                    coro = 1
                    if q_cfg.get("coro"):
                        coro = q_cfg["coro"]
                    for coro_num in range(coro):
                        q_chan = await self.connection.channel()
                        if q_cfg["config"].get("prefetch") is not None:
                            await q_chan.set_qos(prefetch_count=int(q_cfg["config"]["prefetch"]))
                        self.channels.append(q_chan)
                        exchange = await q_chan.declare_exchange(exch_name, **config["config"])
                        queue = await q_chan.declare_queue(
                            q_name, **q_cfg["config"]["arguments"], arguments=q_cfg["config"].get("additional", {})
                        )
                        for bind in q_cfg["bindings"]:
                            logger.info(f"Binding {q_name}->{exch_name}: {bind}")
                            await queue.bind(exchange, bind)
                            # V2 binding
                            if v2_enabled:
                                v2_binding = f"{bind}.{v2_system}"
                                logger.info(f"Binding(v2) {q_name}->{exch_name}: {v2_binding}")
                                await queue.bind(exchange, v2_binding)
                        self.queues[q_name].append(queue)

    async def start(self):
        if self.workers:
            # Run start function
            for w_name, workers in self.workers.items():
                for w in workers:
                    await w.start()
                    logger.info(f"Starting worker {w_name}...")
            self.running = True
            logger.warning(f"Starting {self.name}({self.dispatcher_config['name']})")
            for q_name, channel_list in self.queues.items():
                logger.warning(f"Starting queue {q_name} with {len(channel_list)} coro...")
                coro_num = 0
                for c in channel_list:
                    self.coro.append(self.consumer(q_name, c, coro_num))
                    # self.coro.append(self.consumer_get_method(q_name, c, coro_num))
                    coro_num += 1

    async def consumer_get_method(self, q_name, q_obj, coro_num):
        logger.warning(f"Starting[get] {q_name}[{coro_num}]")
        while self.running:
            try:
                msg = None
                try:
                    msg = await q_obj.get(timeout=2)
                except QueueEmpty:
                    await asyncio.sleep(0.3)
                if self.running and msg:
                    s = time.time()
                    try:
                        await self.proto.dispatch(msg)
                    except Exception:
                        logger.exception(f"Unable to process message for proto: {self.proto}")
                    await msg.ack(multiple=True)
                    logger.info(f"[{coro_num}][get] Message consumed in: {time.time() - s}")
            except Exception:
                logger.exception("[get][{q_name}][{coro_num}] Unable to handle message")
        logger.warning(f"[get]Stopped: {q_name}[{coro_num}]")

    async def consumer(self, q_name, q_obj, coro_num):
        logger.warning(f"Starting {q_name}[{coro_num}]")
        try:
            async with q_obj.iterator() as queue_iter:
                self.queue_iters.append(queue_iter)
                async for message in queue_iter:
                    # Unhandled err
                    try:
                        if self.running:
                            s = time.time()
                            # Return values: MSG_ACK, MSG_NACK, MSG_REJECT
                            msg_action = MSG_ACK
                            try:
                                msg_act = await self.proto.dispatch(message)
                                if msg_act is not None:
                                    msg_action = msg_act
                            except Exception:
                                logger.exception(f"Unable to process message for proto: {self.proto}")
                            if msg_action == MSG_ACK:
                                await message.ack()
                            elif msg_action == MSG_NACK:
                                await message.nack()
                            elif msg_action == MSG_REJECT:
                                await message.reject()
                            logger.info(f"[{coro_num}] Message consumed(FLAG: {msg_action}) in: {time.time() - s}")
                    except (ChannelClosed, ChannelInvalidStateError) as e:
                        logger.error(f"Channel error: {e}. Stopping Shiva!")
                        self.shiva.stop()
        except Exception as err:
            logger.error(err)
        logger.warning(f"Stopped: {q_name}[{coro_num}]")

    async def stop(self):
        self.running = False
        logger.warning("Shutting down queue iters...")
        for i in self.queue_iters:
            logger.info(f"Stopping iter: {i}")
            await i.close()
        logger.warning("Shutting down channels...")
        for ch in self.channels:
            logger.warning(f"Closing channel: {ch}")
            await ch.close()
        logger.warning(f"Shutting down connection {self.connection}...")
        await self.connection.close()
        logger.warning(f"Connection closed {self.connection}...")
        for w_name, w_obj in self.workers.items():
            logger.warning(f"Stopping rmq worker: {w_name}")
            if hasattr(w_obj, "stop"):
                await w_obj.stop()



import prometheus_client
from loguru import logger
from prometheus_client import Gauge, make_asgi_app

from shiva.common.base import BaseDaemon, BaseDispatcher, BaseMetric


class MetricsRoot(BaseDispatcher):
    name = 'metrics_root'
    base_class = BaseMetric

    async def prepare(self):
        self.metrics_app = make_asgi_app()
        if self.dispatcher_config.get('unregister_defaults'):
            prometheus_client.REGISTRY.unregister(prometheus_client.GC_COLLECTOR)
            prometheus_client.REGISTRY.unregister(prometheus_client.PLATFORM_COLLECTOR)
            prometheus_client.REGISTRY.unregister(prometheus_client.PROCESS_COLLECTOR)
        if self.workers:
            logger.info(f'Mounting metrics main: {self.dispatcher_config["uri"]}')
            self.shiva.app.mount(self.dispatcher_config['uri'], self.metrics_app)
        
    async def start(self):
        for d_name, d_list in self.workers.items():
            logger.info(f'Running {d_name}')
            for w in d_list:
                self.coro.append(w.start())

    async def stop(self):
        for d_name, d_list in self.workers.items():
            logger.warning(f'Stopping {d_name}')
            for w in d_list:
                for d in d_list:
                    await d.stop()
        logger.warning(f'Stopped: {self.name}')



import asyncio

from loguru import logger
from prometheus_client import Counter, Gauge

from shiva.common.base import BaseMetric


class MyMetric(BaseMetric):
    name = 'my_metric'
    dispatcher = 'metrics_main'

    async def prepare(self):
        logger.info('Starting main metrics')

    async def start(self):
        logger.info('Metrics started!')
        g = Gauge('my_gauge_metric', 'Some metric')
        c = Counter('my_shit', 'test_counter')
        self.running = True
        while self.running:
            g.inc()
            logger.info('Metrics sleep...')
            c.inc(1.2)
            await asyncio.sleep(5)
    async def stop(self):
        self.running = False



import typer

name = 'maintest'
help_text = 'Shiva tests'
command = typer.Typer()


@command.command()
def test(word: str, second: str = 'default'):
    """
        Test commands
    """
    print(f'WORD: {word}')


@command.command()
def test_1(word: str):
    print(f'WORD: {word}')



import functools
import importlib
import inspect
import os
import pkgutil
from collections import defaultdict

from loguru import logger


def get_class_that_defined_method(meth):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (
        inspect.isbuiltin(meth)
        and getattr(meth, "__self__", None) is not None
        and getattr(meth.__self__, "__class__", None)
    ):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, "__func__", meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth), meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0], None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, "__objclass__", None)  # handle special descriptor objects


class ModuleHelper:
    def __init__(self):
        self.module = None
        self.path = None

    def load_path(self, path):
        try:
            self.module = importlib.import_module(path)
            self.path = path
            return self.module
        except ModuleNotFoundError:
            # Not critical - package/module may not exist
            logger.debug(f"Module path not found (skipping): {path}")
            return None
        except Exception as err:
            # Critical error - module exists but failed to load
            logger.error(f"Unable to import path: {path}: {err}")
            return None

    def load_module(self, module):
        self.module = module
        return self.module

    def get_libs(self, package=None, user_scope=False):
        modules = []
        if not package:
            package = self.module
        submodules = pkgutil.walk_packages(package.__path__, package.__name__ + ".")
        for loader, module_name, is_pkg in submodules:
            # print(f'Module: {module_name}')
            # logger.debug(f'{loader} : {module_name} : {is_pkg}')
            try:
                module = importlib.import_module(module_name)
                if not is_pkg or user_scope:
                    modules.append(module)
            except Exception as e:
                logger.exception(f"Unable to load library: {module_name}: {e}")
        return modules


class Scope:
    def __init__(self, scope_name: str, scopes: list):
        self.scope_name = scope_name
        self.scopes = []
        self.scopes_raw = scopes
        self.load()

    def load(self):
        """
        Load modules from scope paths.

        Note: Duplicates are filtered by module file path.
        First found module wins - later modules with same file are skipped.
        TODO: If override behavior needed (user > packages > shiva),
        reverse the scopes_raw list before loading.
        """
        loaded_files = []
        for scope in self.scopes_raw:
            mh = ModuleHelper()
            success = mh.load_path(scope)
            if success:
                libs = mh.get_libs()
                loaded_count = 0
                for m in libs:
                    if m.__file__ not in loaded_files:
                        self.scopes.append(m)
                        loaded_files.append(m.__file__)
                        loaded_count += 1
                if loaded_count > 0:
                    logger.debug(f"Loaded {loaded_count} modules from: {scope}")
            else:
                # Module path not found - already logged in load_path
                pass

    def filter_members(self, base_class):
        classes = []
        for m in self.scopes:
            members = inspect.getmembers(m, inspect.isclass)
            for member in members:
                action_name, action_class = member
                if issubclass(action_class, base_class) and action_class != base_class:
                    if hasattr(action_class, "load_globals"):
                        # print('---------LOADING--------')
                        for var_name in action_class.load_globals:
                            if hasattr(m, var_name):
                                setattr(action_class, var_name, getattr(m, var_name))
                    classes.append(action_class)
        return classes

    def filter_objects(self, base_class):
        classes = []
        for m in self.scopes:
            members = inspect.getmembers(m, inspect.ismemberdescriptor)
            for member in members:
                action_name, action_class = member
                # print(f'{action_name}: {action_class}')
                if issubclass(action_class, base_class) and action_class != base_class:
                    classes.append(action_class)
                #     if action_class.dispatcher in self.dispatchers.values():
                #         logger.info(f'Action "{action_class}" imported!')
                #         self.actions[action_class.dispatcher.name].append(action_class)
        return classes



from loguru import logger
from shiva.common.base import BaseDaemon, BaseDispatcher


class DaemonRoot(BaseDispatcher):
    name = 'daemon_root'
    base_class = BaseDaemon

    async def prepare(self):
        # print('*' * 80)
        # print(self.workers_configurations)
        # print(self.workers_all)
        # print(self.workers)
        # print('*' * 80)
        pass
        
    async def start(self):
        for d_name, d_list in self.workers.items():
            logger.info(f'Running {d_name}')
            for w in d_list:
                self.coro.append(w.start())

    async def stop(self):
        for d_name, d_list in self.workers.items():
            logger.warning(f'Stopping {d_name}')
            for w in d_list:
                for d in d_list:
                    await d.stop()
        logger.warning(f'Stopped: {self.name}')
